<?php

namespace Drupal\search_api_aais\Azure\Query;

use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\search_api\Query\QueryInterface;
use Drupal\search_api\Utility\FieldsHelperInterface;
use Drupal\search_api_aais\Azure\Index\IndexBuilder;

/**
 * Query parameter builder.
 */
class QueryParamBuilder implements QueryParamBuilderInterface {

  /**
   * The query filter builder.
   *
   * @var \Drupal\search_api_aais\Azure\Query\QueryFilterBuilder
   */
  protected QueryFilterBuilder $queryFilterBuilder;

  /**
   * The query sort builder.
   *
   * @var \Drupal\search_api_aais\Azure\Query\QuerySortBuilder
   */
  protected QuerySortBuilder $querySortBuilder;

  /**
   * The Search API fields helper.
   *
   * @var \Drupal\search_api\Utility\FieldsHelperInterface
   */
  protected FieldsHelperInterface $fieldsHelper;

  /**
   * The language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected LanguageManagerInterface $languageManager;

  /**
   * The constructor.
   *
   * @param \Drupal\search_api_aais\Azure\Query\QueryFilterBuilder $query_filter_builder
   *   The query filter builder.
   * @param \Drupal\search_api_aais\Azure\Query\QuerySortBuilder $query_sort_builder
   *   The query sort builder.
   * @param \Drupal\search_api\Utility\FieldsHelperInterface $fields_helper
   *   The Search API fields helper.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   */
  public function __construct(QueryFilterBuilder $query_filter_builder, QuerySortBuilder $query_sort_builder, FieldsHelperInterface $fields_helper, LanguageManagerInterface $language_manager) {
    $this->queryFilterBuilder = $query_filter_builder;
    $this->querySortBuilder = $query_sort_builder;
    $this->fieldsHelper = $fields_helper;
    $this->languageManager = $language_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function buildQueryParams(QueryInterface $query): array {
    $index = $query->getIndex();

    /** @var \Drupal\search_api_aais\AzureAiSearchBackendInterface $backend */
    $backend = $index->getServerInstance()->getBackend();

    // Set-up some default parameters for a semantic search.
    $params = [
      // (false)|true.
      'count' => TRUE,
      // (any)|all.
      'searchMode' => 'any',
      // (*)|..
      'select' => 'itemId',
      // (disabled)|enabled
      'featuresMode' => 'enabled',
      // (simple)|full|semantic
      'queryType' => 'semantic',
      // (none)|extractive
      'answers' => 'extractive',
      // (50)|0..1000
      'top' => $query->getOption('limit', 10),
      // (0)|0..100000
      'skip' => $query->getOption('offset', 0),
      // (none)|lexicon
      'speller' => 'lexicon',
    ];

    // If the configured API-version in the backend requires semantic
    // configuration, pass it on.
    if ($backend->requiresSemanticConfiguration()) {
      $params['semanticConfiguration'] = IndexBuilder::getIndexConfiguration($index, 'semantic_configuration');
    }

    // Set all full-text fields as the fields to actually search in.
    $params['searchFields'] = implode(', ', $index->getFulltextFields());

    // Check for any language filter(s) that needs to be applied.
    if ($language_filters = $query->getLanguages()) {
      // If we need to filter on one or more languages, add it to the conditions
      // so that it will be picked up by the query filter builder.
      // @see \Drupal\search_api_aais\Azure\Query\QueryFilterBuilder
      $query->getConditionGroup()->addCondition('search_api_language', $language_filters, 'IN');

      // Set the query language based on the language filters.
      // However, Azure only accepts a single language as a filter, so we just
      // take the first one.
      $query_language = reset($language_filters);
    }
    else {
      // If there are no language filters, set the query language based on the
      // current language.
      $query_language = $this->languageManager->getCurrentLanguage()->getId();
    }

    // Set the query language.
    $params['queryLanguage'] = $query_language;

    // Construct the filters.
    $filters = $this->queryFilterBuilder->buildQueryFilters($query->getConditionGroup(), IndexBuilder::getIndexFields($index));

    if (!empty($filters)) {
      // @todo Improve the hardcoded 'and' concatenation.
      $params['filter'] = implode(' and ', $filters);
    }

    // Construct the sorts.
    $sorts = $this->querySortBuilder->buildQuerySorts($query->getSorts(), IndexBuilder::getIndexFields($index));

    if (!empty($sorts)) {
      $params['orderby'] = implode(', ', $sorts);

      // Sorting cannot be applied if the queryType is 'semantic', as Azure then
      // automatically sorts the results based on their relevance.
      // So if any sorting is present on a 'semantic' queryType, switch to the
      // 'full' queryType instead and change the parameters as needed.
      if ($params['queryType'] === 'semantic') {
        $params['queryType'] = 'full';
        unset($params['answers']);
        unset($params['semanticConfiguration']);
      }
    }

    // If the captions option is enabled and the queryType is 'semantic', apply
    // the configured highlight tags.
    if ($params['queryType'] === 'semantic' && IndexBuilder::getIndexConfiguration($index, 'captions')) {
      $tag = IndexBuilder::getIndexConfiguration($index, 'highlight_tag') ?? 'em';

      $params['captions'] = 'extractive';
      $params['highlightPreTag'] = sprintf('<%s>', $tag);
      $params['highlightPostTag'] = sprintf('</%s>', $tag);
    }

    // Set the search keys as the actual search query.
    if ($keys = $query->getKeys()) {
      $keys = array_filter($keys, function (string $key) {
        return $key[0] !== '#';
      }, ARRAY_FILTER_USE_KEY);

      $params['search'] = implode(' ', $keys);
    }

    return $params;
  }

}
